package org.bitbucket.bradleysmithllc.etlunit.feature.informatica;

import org.apache.commons.io.FileUtils;
import org.bitbucket.bradleysmithllc.etlagent.informatica.dto.RemoteFile;
import org.bitbucket.bradleysmithllc.etlagent.informatica.provider.*;
import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.feature.Feature;
import org.bitbucket.bradleysmithllc.etlunit.feature.RuntimeOption;
import org.bitbucket.bradleysmithllc.etlunit.feature.extend.Extender;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.FileFeatureModule;
import org.bitbucket.bradleysmithllc.etlunit.feature.informatica.json.informatica.execute.ExecuteHandler;
import org.bitbucket.bradleysmithllc.etlunit.feature.logging.LogFileManager;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestOperation;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObjectBuilder;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObjectImpl;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;
import org.bitbucket.bradleysmithllc.etlunit.util.VelocityUtil;
import org.bitbucket.bradleysmithllc.webserviceshubclient.parameter.ParameterFile;

import javax.inject.Inject;
import javax.inject.Named;
import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class InformaticaExecutor implements Extender, ExecuteHandler {
	public static final String INFORMATICA_CREATED_FOLDERS = "createdFolders";
	public static final String INFORMATICA_LOADED_WORKFLOWS = "loadedWorkflows";
	public static final String INFORMATICA_EXPORTED_WORKFLOWS = "exportedWorkflows";
	public static final String INFORMATICA_CREATED_CONNECTIONS = "createdConnections";

	private Log applicationLog;
	private final InformaticaFeatureModule informaticaFeature;
	private final File informaticaResourceDir;
	private final String projectIdentifier;
	private RuntimeSupport runtimeSupport;
	private boolean loadWorkflows = true;
	private boolean exportWorkflows = false;
	private boolean prepareWorkspace = true;
	private LogFileManager logFileManager;

	private InformaticaRuntimeSupport informaticaRuntimeSupport;
	private static final String INFORMATICA_CONTEXTS = "INFORMATICA_CONTEXTS";

	public InformaticaExecutor(InformaticaFeatureModule informaticaFeature, File informaticaResourceDir, String projectIdentifier) {
		this.informaticaFeature = informaticaFeature;
		this.informaticaResourceDir = informaticaResourceDir;
		this.projectIdentifier = projectIdentifier;
	}

	@Inject
	public void setInformaticaRuntimeSupport(InformaticaRuntimeSupport support) {
		informaticaRuntimeSupport = support;
	}

	@Inject
	public void setLogFileManager(LogFileManager logFileManager) {
		this.logFileManager = logFileManager;
	}

	@Inject
	public void setLoadWorkflows(@Named("informatica.loadWorkflows") RuntimeOption option) {
		loadWorkflows = option.isEnabled();

		applicationLog.info(loadWorkflows ? "Loading workflows from project" : "Using workflows resident in informatica repository");
	}

	@Inject
	public void setPrepareWorkspace(@Named("informatica.prepareWorkspace") RuntimeOption option) {
		prepareWorkspace = option.isEnabled();

		applicationLog.info(prepareWorkspace ? "Preparing database" : "Using workflows resident in informatica repository");
	}

	@Inject
	public void setExportWorkflows(@Named("informatica.exportWorkflows") RuntimeOption option) {
		exportWorkflows = option.isEnabled();

		applicationLog.info(exportWorkflows ? "Updating local workflows" : "Not exporting workflows");
	}

	@Inject
	public void setLogger(@Named("applicationLog") Log log) {
		applicationLog = log;
	}

	@Inject
	public void setRuntimeSupport(RuntimeSupport runtimeSupport) {
		this.runtimeSupport = runtimeSupport;
	}

	public boolean canHandleRequest(ETLTestOperation etlTestOperation, ETLTestValueObject etlTestValueObject, VariableContext context, ExecutionContext executionContext) {
		return
				etlTestValueObject != null
						&&
						etlTestValueObject.query("folder") != null
						&&
						etlTestValueObject.query("workflow") != null;
	}

	public ClassResponder.action_code process(ETLTestOperation etlTestOperation, ETLTestValueObject etlTestValueObject, VariableContext context, ExecutionContext executionContext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		return action_code.defer;
	}

	public Feature getFeature() {
		return informaticaFeature;
	}

	private void declareConnectionsIfNeeded(org.bitbucket.bradleysmithllc.etlunit.feature.informatica.json.informatica.execute.ExecuteRequest operation, ETLTestValueObject etlTestValueObject, ETLTestValueObject infaConnections, VariableContext context) throws TestExecutionError {
		VariableContext topper = context.getTopLevelScope();

		// search for the database.connections path
		if (!context.hasVariableBeenDeclared("database")) {
			// nothing to do
			return;
		}

		ETLTestValueObject database = context.getValue("database");

		if (database != null) {
			ETLTestValueObject connections = database.query("connections");

			if (connections != null) {
				InformaticaRepository ipc = informaticaFeature.repositoryForRequest(etlTestValueObject);

				// iterate through, and create connections in informatica as required
				for (Map.Entry<String, ETLTestValueObject> connEntry : connections.getValueAsMap().entrySet()) {
					// each element here is a connection id mapped to a mode
					String connId = connEntry.getKey();
					ETLTestValueObject value = connEntry.getValue();

					for (Map.Entry<String, ETLTestValueObject> modeEntry : value.getValueAsMap().entrySet()) {
						// here the id is the mode name.  This can be null if it is 'modeless'
						String mode = modeEntry.getKey();
						ETLTestValueObject modeValue = modeEntry.getValue();

						// grab the required information
						String dbname = modeValue.query("database-name").getValueAsString();
						String login = modeValue.query("login-name").getValueAsString();
						String password = modeValue.query("password").getValueAsString();
						String server = modeValue.query("server-name").getValueAsString();

						String variableName = "informatica_connections_" + connId + "_" + mode;
						if (infaConnections.query(variableName) == null) {
							applicationLog.info("Creating informatica connection for id = '" + connId + "', mode = '" + mode + "', databaseName = '" + dbname + "', login = '" + login + "', password = '" + password + "', server-name = '" + server + "'");
							new ETLTestValueObjectBuilder(infaConnections).key(variableName).value("true");

							// execute informatica create statement, and place a context variable pointing
							// to the connection name
							// let the database name take care of scoping project and version, we will prepend the user name
							String connectionName = new StringBuilder(runtimeSupport.getProjectUser()).append("@").append(connId).append("_").append(mode).append("_").append(projectIdentifier).toString();

							try {
								// create the connection, and store in the variable context  This is hard-coded for sqlserver
								// now, so it will need to be changed when a second database is introduced
								InetAddress serverA = InetAddress.getByName(server);
								InformaticaRepositoryClient repoclient = ipc.getInformaticaRepositoryClient();

								// first delete the connection, then create.  Do not fail if the connection does not exist
								repoclient.deleteConnection(connectionName, false);
								repoclient.createConnection(serverA.getHostAddress(), dbname, login, password, connectionName);

								// this is dicey since some might have already been created
								Map<String, ETLTestValueObject> databaseMap;
								Map<String, ETLTestValueObject> connectionsMap;

								if (!topper.hasVariableBeenDeclared("informatica")) {
									databaseMap = new HashMap<String, ETLTestValueObject>();

									ETLTestValueObject obj = new ETLTestValueObjectImpl(databaseMap);

									topper.declareAndSetValue("informatica", obj);

									// create the variable for connections
									connectionsMap = new HashMap<String, ETLTestValueObject>();
									obj = new ETLTestValueObjectImpl(connectionsMap);

									databaseMap.put("connections", obj);
								} else {
									ETLTestValueObject informatica = topper.getValue("informatica");
									connectionsMap = informatica.query("connections").getValueAsMap();
								}

								// look for this connection id
								Map<String, ETLTestValueObject> connectionMap;
								if (connectionsMap.containsKey(connId)) {
									connectionMap = connectionsMap.get(connId).getValueAsMap();
								} else {
									connectionMap = new HashMap<String, ETLTestValueObject>();
									ETLTestValueObject obj = new ETLTestValueObjectImpl(connectionMap);
									connectionsMap.put(connId, obj);
								}

								connectionMap.put(mode, new ETLTestValueObjectImpl(connectionName));
							} catch (Exception e) {
								applicationLog.severe("Error creating relational connection", e);
								throw new RuntimeException();
							}
						}
					}
				}
			}
		}

		// at this point, check for a context request.  If found, we need to create all variables to match the requested ones
		Map<String, String> map = operation.getContext().getAdditionalProperties();

		if (map != null) {
			for (Map.Entry<String, String> entry : map.entrySet()) {
				// This is a simple aliasing process, and a bit of a workaround.
				// the name is a context variable name, the value is another variable name to set the 'name' context variable to.
				String name = entry.getKey();
				String value = entry.getValue();

				applicationLog.info("Aliasing [" + name + "] to [" + value + "]");
				context.declareAndSetValue(name, context.query(value));
			}
		}
	}

	public action_code execute(final org.bitbucket.bradleysmithllc.etlunit.feature.informatica.json.informatica.execute.ExecuteRequest operation, final ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning {
		InformaticaIntegrationService integrationService = informaticaFeature.integrationServiceForRequest(obj);
		InformaticaRepository informaticaRepository = informaticaFeature.repositoryForRequest(obj);

		InformaticaDomain informaticaDomain = integrationService.getInformaticaDomain();

		// grab the domain object
		ETLTestValueObject container = getInformaticaContainer(context, true);

		if (container.query(informaticaDomain.getDomainName()) == null) {
			new ETLTestValueObjectBuilder(container).key(informaticaDomain.getDomainName()).value(new ETLTestValueObjectImpl(new HashMap()));
		}

		ETLTestValueObject domainContainer = container.query(informaticaDomain.getDomainName());
		ETLTestValueObject repositoryContainer = null;

		// do the same for the repositories
		if (domainContainer.query(informaticaRepository.getRepositoryName()) == null) {
			new ETLTestValueObjectBuilder(domainContainer).key(informaticaRepository.getRepositoryName()).value(new ETLTestValueObjectImpl(new HashMap()));

			// create the list for created folders, workflows, etc.
			repositoryContainer = domainContainer.query(informaticaRepository.getRepositoryName());
			new ETLTestValueObjectBuilder(repositoryContainer)
					.key(INFORMATICA_CREATED_FOLDERS)
					.value(new ETLTestValueObjectImpl(new HashMap<String, ETLTestValueObject>()))
					.key(INFORMATICA_LOADED_WORKFLOWS)
					.value(new ETLTestValueObjectImpl(new HashMap<String, ETLTestValueObject>()))
					.key(INFORMATICA_EXPORTED_WORKFLOWS)
					.value(new ETLTestValueObjectImpl(new HashMap<String, ETLTestValueObject>()))
					.key(INFORMATICA_CREATED_CONNECTIONS)
					.value(new ETLTestValueObjectImpl(new HashMap<String, ETLTestValueObject>()));
		} else {
			repositoryContainer = domainContainer.query(informaticaRepository.getRepositoryName());
		}

		// handle checking for connections here, since the method-level broadcast might be in any order
		// since these connections are created by the database module and there is no explicit dependency
		// between these two
		declareConnectionsIfNeeded(operation, obj, repositoryContainer.query(INFORMATICA_CREATED_CONNECTIONS), context);

		String folder = operation.getFolder();
		String workflow = operation.getWorkflow();

		// determine where the param file and the workflow export file are
		File xmlFile = informaticaRuntimeSupport.getInformaticaWorkflow(folder, workflow);

		File paramFileTemplate = informaticaRuntimeSupport.getInformaticaWorkflowParameterTemplate(folder, workflow);

		if (!paramFileTemplate.exists()) {
			try {
				IOUtils.writeBufferToFile(paramFileTemplate, new StringBuffer("[Global]"));
			} catch (IOException e) {
				throw new TestExecutionError("Error creating parameter file template", InformaticaConstants.ERR_CREATE_PRM_FILE, e);
			}
		}

		// set standard variables -
		File srcFiles = informaticaRuntimeSupport.getSrcFilesRemote(informaticaDomain);
		File lkpFiles = informaticaRuntimeSupport.getLkpFilesRemote(informaticaDomain);
		File tgtFiles = informaticaRuntimeSupport.getTgtFilesRemote(informaticaDomain);
		File badFiles = informaticaRuntimeSupport.getBadFilesRemote(informaticaDomain);
		File logFiles = informaticaRuntimeSupport.getLogFilesRemote(informaticaDomain);
		File parmlib = informaticaRuntimeSupport.getParmFilesRemote(informaticaDomain);

		// process staged files
		// look in either the default or user-specified context, and any files listed move them into the
		// appropriate location
		ETLTestValueObject fileContext = FileFeatureModule.getFileContextContainer(obj, context);

		Map<RemoteFile.fileType, List<RemoteFile>> remotes = new HashMap<RemoteFile.fileType, List<RemoteFile>>();

		if (fileContext != null) {
			// in order to facilitate downstream file operations, this module
			// needs to store locations for our classifiers in the file context
			// so that these paths are available to the other processes.
			// if the destination-classifiers has already been defined, then do nothing
			if (fileContext.query("destination-classifiers") == null) {
				ETLTestValueObject builtObject = new ETLTestValueObjectBuilder(fileContext)
						.key("destination-classifiers")
						.object()
						.key("source")
						.value(srcFiles.getAbsolutePath())
						.key("lookup")
						.value(lkpFiles.getAbsolutePath())
						.key("target")
						.value(tgtFiles.getAbsolutePath())
						.endObject()
						.toObject();
			}

			// populate the file map for the workflow
			fileContext = fileContext.query("file-list");

			// process files to the working root by classification.  source is default
			for (ETLTestValueObject object : fileContext.getValueAsList()) {
				String filePath = object.query("file").getValueAsString();

				String classifier = null;
				ETLTestValueObject classifierObj = object.query("classifier");

				if (classifierObj != null) {
					classifier = classifierObj.getValueAsString();
				}

				File source = new File(filePath);
				String destName = source.getName();
				ETLTestValueObject destNameObj = object.query("destination-name");

				if (destNameObj != null) {
					destName = destNameObj.getValueAsString();
				}

				File destDir = informaticaRuntimeSupport.getSrcFilesRemote(informaticaDomain);

				// post the sucker
				try {
					RemoteFile rf = null;

					if (classifier == null || classifier.equals("source")) {
						// nothing to do - just here for validation
						rf = RemoteFile.newRemoteFile(source, destName, RemoteFile.fileType.sourceFiles);
					} else if (classifier.equals("target")) {
						rf = RemoteFile.newRemoteFile(source, destName, RemoteFile.fileType.targetFiles);
					} else if (classifier.equals("lookup")) {
						rf = RemoteFile.newRemoteFile(source, destName, RemoteFile.fileType.lookupFiles);
					} else {
						throw new TestExecutionError("Invalid file classification: '" + classifier + "'.  Classifier must be one of [source, target, lookup].", InformaticaConstants.ERR_INVALID_FILE_CLASSIFIER);
					}

					List<RemoteFile> remoteFileList = remotes.get(rf.getType());

					if (remoteFileList == null) {
						remoteFileList = new ArrayList<RemoteFile>();
						remotes.put(rf.getType(), remoteFileList);
					}

					remoteFileList.add(rf);
				} catch (IOException e) {
					throw new TestExecutionError("Error copying stage file to remote: source = '" + source + "'", InformaticaConstants.ERR_STAGE_FILE, e);
				}
			}
		}


		try {
			RemoteFile.copyRemoteToLocal(remotes, informaticaDomain);
		} catch (IOException e) {
			throw new TestExecutionError("", e);
		}

		File parm = new File(parmlib, paramFileTemplate.getName());

		try {
			String prm = IOUtils.readFileToString(paramFileTemplate);

			IOUtils.writeBufferToFile(parm, new StringBuffer(VelocityUtil.writeTemplate(prm, context, informaticaRuntimeSupport.getInformaticaWorkflowParameterDirectory(folder))));

			applicationLog.info("Created parm file " + parm.getAbsolutePath());
		} catch (Exception e) {
			throw new TestExecutionError("Error creating parameter file", InformaticaConstants.ERR_CREATE_PRM_FILE, e);
		}

		applicationLog.info("Running workflow " + workflow + " using template " + paramFileTemplate.getAbsolutePath() + " and xml definition " + xmlFile.getAbsolutePath());

		String projectGlom = runtimeSupport.getProjectUser() + "_" + projectIdentifier;

		// grab the project version
		try {
			String realFolder = folder;

			if (loadWorkflows) {
				InformaticaRepositoryClient repoClient = informaticaRepository.getInformaticaRepositoryClient();

				realFolder = "__" + projectGlom + "_" + folder;

				// load the folders as needed
				List<String> flist = InformaticaRepositoryClientImpl.getFolderList(xmlFile);

				ETLTestValueObject folderList = repositoryContainer.query(INFORMATICA_CREATED_FOLDERS);

				for (String ifolder : flist) {
					String folderName = "__" + projectGlom + "_" + ifolder;

					// check to see if this has been created yet
					if (folderList.query(folderName) == null) {
						// drop and create folder - don't care if it doesn't exist
						repoClient.deleteFolder(folderName, false);
						repoClient.createFolder(folderName);
						new ETLTestValueObjectBuilder(folderList).key(folderName).value("true");
					}
				}

				ETLTestValueObject workflowList = repositoryContainer.query(INFORMATICA_LOADED_WORKFLOWS);

				// load the workflow into the repository
				// check to see if this has been loaded yet
				String workName = folder + "_______" + workflow;

				if (workflowList.query(workName) == null) {
					repoClient.importWorkflowFromXml("__" + projectGlom, integrationService, xmlFile);
					new ETLTestValueObjectBuilder(workflowList).key(workName).value("true");
				}
			}

			if (exportWorkflows) {
				InformaticaRepositoryClient repoClient = informaticaRepository.getInformaticaRepositoryClient();

				String key = "[[Exported[" + folder + "/" + workflow + "]]]";

				if (!context.getTopLevelScope().hasVariableBeenDeclared(key)) {
					applicationLog.info("Exporting " + folder + "/" + workflow);
					repoClient.exportWorkflowToXml(workflow, folder, xmlFile);
					context.getTopLevelScope().declareVariable(key);
				}
			}

			String task = operation.getTask();

			InformaticaExecutionClient informaticaIntegrationServiceClient = informaticaFeature.getIntegrationServiceClient(integrationService);

			ParameterFile pmf = new ParameterFile(parm);

			Map<RemoteFile.fileType, List<RemoteFile>> responseRemotes = null;

			if (task != null) {
				applicationLog.info("Executing task " + realFolder + "/" + workflow + "." + task);

				responseRemotes = informaticaIntegrationServiceClient.executeWorkflowTask(workflow, task, realFolder, pmf, remotes);
			} else {
				applicationLog.info("Executing workflow " + realFolder + "/" + workflow);

				responseRemotes = informaticaIntegrationServiceClient.executeWorkflow(workflow, realFolder, pmf, remotes);
			}

			String logName = operation.getLogName();

			if (logName == null) {
				logName = operation.getWorkflow() + (operation.getTask() != null ? ("." + operation.getTask()) : "");
			}

			if (responseRemotes != null)
			{
				for (Map.Entry<RemoteFile.fileType, List<RemoteFile>> remoteEntry : responseRemotes.entrySet())
				{
					RemoteFile.fileType type = remoteEntry.getKey();
					List<RemoteFile> remoteFileList = remoteEntry.getValue();

					for (RemoteFile remoteFile : remoteFileList)
					{
						File destFile = new File(informaticaDomain.getWorkspaceDir(type), remoteFile.getFileName());

						applicationLog.info("Caching log file: " + remoteFile + " to " + destFile);

						FileUtils.write(destFile, remoteFile.getFileContents());

						logFileManager.addLogFile(op, destFile, type.name(), logName);
					}
				}
			}
		} catch (Exception e) {
			throw new TestExecutionError("Error executing workflow", InformaticaConstants.ERR_EXECUTE_WORKFLOW, e);
		}

		return ClassResponder.action_code.handled;
	}

	public static ETLTestValueObject getInformaticaContainer(VariableContext context, boolean create) {
		VariableContext topper = context.getTopLevelScope();

		if (!topper.hasVariableBeenDeclared(INFORMATICA_CONTEXTS)) {
			if (create) {
				topper.declareAndSetValue(INFORMATICA_CONTEXTS, new ETLTestValueObjectImpl(new HashMap()));
			} else {
				return null;
			}
		}

		return topper.getValue(INFORMATICA_CONTEXTS);
	}
}